# -*- coding: UTF-8 -*-
"""Plot OSIRIS-generated data."""

from __future__ import print_function
from __future__ import division

from glob import glob
import os
import re

import h5py
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import ticker
from matplotlib.animation import FuncAnimation, FFMpegWriter

from ..common import ensure_dir_exists, human_order_key, MPCaller, Call, logger


def _is_latex(s):
    """

    Decide if a string is a LaTeX expression.

    Args:
        s (str): The string to test.

    Returns: (bool) True if it seems a LaTeX expression, False otherwise.

    """
    if not s:  # Empty string is not LaTeX
        return False
    return re.match("^.(_.)?$", s) or "\\" in s  # If in the form x, x_i, or if there is a backlash (command)


class Diagnostic:
    """
    A OSIRIS diagnostic.

    Attributes:
        data_path (str): Path to the directory were the data is stored.
        data_name (str): A friendly name for the data.
        units (str): The name of the unit the magnitude is measured in.
        dt (float): The time step between snapshots of a consecutive number.
        t_0 (float): The time of the first snapshot.
        time_units (str): The name of the unit of time.
        file_list (`list` of `str`): List of h5 files, one per time snapshot.
        snapshot_list (`list` of `int`): List of integers identifying the snapshots. Multiply by dt to get time.
        keys (`list` of `str`): Names of the datasets in the Diagnostic, given in human order.
        axes (`list` of `dict`): Info of each axis in the Diagnostic.
        datasets_as_axis(`dict`): Info of datasets if treated as axes. WARNING: Only used with energy bins.
        shape (`tuple`): A tuple with:

            * `list`: The number of grid dimensions.
            * `int`: The number of datasets (excluding axes definition).
            * `int`: The number of snapshots in time.
            
    Note: The axes list is provided in the order of the numpy convention for arrays. This is the opposite of order used
        to label the axes in the hdf5 files. For example in a 2d array the first axes will be the labeled as AXIS2, and
        the second will be AXIS1. Unless the user makes use of other external tools to read the data, he/she can safely
        ignore this note.
            
    """

    def __init__(self, data_path):
        """
        Create a Diagnostic instance.

        Args:
            data_path: Path of the directory containing the diagnostic data.
            
        """
        self.data_path = data_path
        self.file_list = glob(os.path.join(data_path, "*.h5"))
        if not self.file_list:
            raise ValueError("No diagnostic data found in %s" % data_path)
        self.file_list.sort(key=human_order_key)
        self.snapshot_list = list(
            map(lambda x: int((os.path.split(x)[1]).split(".h5")[0].split("-")[-1]), self.file_list))

        # Get info from first time snapshot
        with h5py.File(self.file_list[0], "r") as f:
            self.t_0 = f.attrs["TIME"][0]
            self.time_units = f.attrs["TIME UNITS"][0].decode('UTF-8')
            self.keys = self._get_keys(f)
            self.axes = self._get_axes(f, self.keys[0])
            self.units = f[self.keys[0]].attrs["UNITS"][0].decode('UTF-8')
            if len(self.keys) > 1:
                # Take a general name from the global attribute
                self.data_name = f.attrs["NAME"][0].decode('UTF-8')
                # Construct an axes-like object representing the dataset
                self.datasets_as_axis = {}
                # Guess name
                name = f[self.keys[0]].attrs["TAG"][0].decode('UTF-8')  # E.g., "Energy <=    0.00000000"
                name = re.match("(.*?) <", name).group(1)  # TODO: Catch errors, check syntax
                self.datasets_as_axis["NAME"] = name
                self.datasets_as_axis["LONG_NAME"] = name
                # Guess values
                dataset_axes = []
                number_pattern = r'[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?'  # A pattern matching floats
                for d in self.keys:
                    tag = f[d].attrs["TAG"][0].decode('UTF-8')
                    # tag is, e.g., '   0.00000000      < Energy <=    5.00000007E-02'
                    values = list(map(float, re.findall(number_pattern, tag)))
                    dataset_axes.append(sum(values) / len(values))  # So simple inequalities are represented somehow
                    # TODO: First and last datasets are not bounded intervals. Their only finite point is used as the x coordinate, which must be understood with caution
                self.datasets_as_axis["LIST"] = dataset_axes
                self.datasets_as_axis["MIN"] = dataset_axes[0]
                self.datasets_as_axis["MAX"] = dataset_axes[-1]
                self.datasets_as_axis["UNITS"] = "?"
            else:
                # Take a specific name from the dataset
                self.data_name = f[self.keys[0]].attrs["LONG_NAME"][0].decode('UTF-8')
                # No axes-like object in this case
                self.datasets_as_axis = None

        if len(self.file_list) < 2:
            self.dt = 0
        else:
            with h5py.File(self.file_list[1], "r") as f:
                self.dt = f.attrs["TIME"][0] / self.snapshot_list[1] - self.t_0

        self.shape = ([len(x["LIST"]) for x in self.axes], len(self.keys), len(self.snapshot_list))

    def __repr__(self):
        return "Diagnostic<%s %s>" % (self.data_name, str(self.shape))

    @staticmethod
    def _get_keys(file):
        """Get the dataset keys from an opened file."""
        keys = list(file.keys())
        if "AXIS" not in keys:
            raise ValueError("AXIS group not found.")
        keys.remove("AXIS")
        keys.sort(key=human_order_key)
        return keys

    @staticmethod
    def _get_axes(file, dataset_key=None):
        """Get the axes info."""
        if dataset_key is None:
            dataset_key = Diagnostic._get_keys(file)[0]
        axes = []
        for i, axis in enumerate(reversed(list(file["AXIS"].keys()))):
            ax = file["AXIS"][axis]
            data = {}
            for d in ["LONG_NAME", "UNITS", "NAME", "TYPE"]:
                data[d] = ax.attrs[d][0].decode('UTF-8')
            data["MIN"], data["MAX"] = ax[:]
            # TODO: Non linear axis support
            data["LIST"] = np.linspace(data["MIN"], data["MAX"], num=file[dataset_key].shape[i])
            axes.append(data)

        return axes

    def _clean_dataset_key(self, dataset_key):
        """Return the given dataset key as str, using human order if int. Might raise error or warning."""
        if isinstance(dataset_key, int):
            dataset_key = self.keys[dataset_key]
        elif isinstance(dataset_key, str):
            if dataset_key not in self.keys:
                raise ValueError("Dataset %s does not exist in the file." % dataset_key)
        elif dataset_key is None:
            if len(self.keys) != 1:  # Warn if implicitly selecting one among others.
                logger.warning("No dataset selected when multiple are available. Plotting the first one.")
            dataset_key = self.keys[0]
        else:
            raise TypeError("Unknown dataset type: %s", type(dataset_key))
        return dataset_key

    def get_generator(self, dataset_selector=None, axes_selector=None, time_selector=None):
        """
        Get a generator providing data from the file.

        Calling this method returns a generator which, when called, will provide data for increasing times (unless
        modified by time_selector parameter). The data might be reduced either by selecting a position in an axis (or
        a dataset) or by using a function along some axis (or datasets), e.g., a sum.

        This data is provided as numpy arrays where the first axis refers to dataset coordinate (if present) and next
        to (non-reduced) axis in the order they are found in the files.

        Args:
            dataset_selector (str, int or callable): Instructions to reduce datasets. An int selects a dataset in human
                order, a str selects it by name. A function taking a list and returning a scalar can be used to reduce
                the data, e.g., sum, mean...

            axes_selector (tuple): Instructions to reduce axes data. It must be
                a tuple of the same length of the number axes or None to perform no reduction.
                Each element can be of the following types:

                    * int: Select the item in the given position.
                    * None: No reduction is performed in this axis.
                    * callable (default): Reduce the data along this axes using the given function (e.g., mean, max, sum...).


            time_selector (slice): A slice instance selecting the points in time to take.

        Returns: (generator): A generator which provides the data.

        """
        multiple_datasets = False  # If a dataset list is going to be returned
        if dataset_selector is not None:
            if self.shape[1] == 1:
                logger.warning("Single dataset found. Ignoring the provided dataset_selector.")

                def f_dataset_selector(f):
                    return f[self.keys[0]][:]
            else:
                if isinstance(dataset_selector, int):
                    dataset_selector = self.keys[dataset_selector]

                if isinstance(dataset_selector, str):  # If it was int or str
                    def f_dataset_selector(f):
                        return f[dataset_selector][:]
                else:  # Assumed function
                    def f_dataset_selector(f):
                        return np.apply_along_axis(dataset_selector, 0, [f[key][:] for key in self.keys])

        else:
            if self.shape[1] > 1:
                multiple_datasets = True

                def f_dataset_selector(f):
                    return np.array([f[key][:] for key in self.keys])
            else:
                def f_dataset_selector(f):
                    return f[self.keys[0]][:]

        if axes_selector is not None:
            def f_axes_selector(x):
                offset = 1 if multiple_datasets else 0  # If multiple dataset, do not count its axis for reduction
                for i, sel in enumerate(axes_selector):
                    if sel is not None:
                        if isinstance(sel, int):
                            x = np.take(x, sel, axis=i - offset)
                        else:  # Assumed function
                            x = np.apply_along_axis(sel, i - offset, x)
                        offset += 1
                return x
        else:
            def f_axes_selector(x):
                return x

        if time_selector is not None and not isinstance(time_selector, slice):
            logger.warning("Invalid time_selector parameter ignored. Use a slice instead.")
            time_selector = None

        def gen():
            for file_name in (self.file_list[time_selector] if time_selector else self.file_list):
                with h5py.File(file_name, "r") as f:
                    data = f_dataset_selector(f)
                # Make sure to exit the context manager before yielding
                # h5py might accuse you of murdering identifiers if you don't!
                yield f_axes_selector(data)

        return gen()

    def get_axes(self, dataset_selector=None, axes_selector=None):
        """
        Get a dictionary with the info of the axes obtained as result of a given reduction.
        
        Args:
            dataset_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            axes_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.

        Returns:
            (list of dict) Ordered list of the axes left by the reduction.
            
        """
        axes = []
        if dataset_selector is not None:
            if self.shape[1] == 1:
                logger.warning("Single dataset found. Ignoring the provided dataset_selector.")
                # Anyhow, datasets are reduced, so skip

        elif self.shape[1] > 1:
            axes.append(self.datasets_as_axis)

        if axes_selector is not None:
            for i, sel in enumerate(axes_selector):
                if sel is None:
                    axes.append(self.axes[i])
        else:
            for a in self.axes:
                axes.append(a)
        return axes

    def get_time_list(self, time_selector=None):
        """
        Get the list of times obtained as a result of a given slice.
        
        Args:
            time_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method. 

        Returns:

        """
        if time_selector:
            # This could be improved to avoid generating unneeded values
            return [self.t_0 + self.dt * i for i in self.snapshot_list][time_selector]
        else:
            return [self.t_0 + self.dt * i for i in self.snapshot_list]

    def time_1d_animation(self, output_path=None, dataset_selector=None, axes_selector=None, time_selector=None,
                          dpi=200, fps=1, scale_mode="expand",
                          latex_label=True):
        """
        Generate a plot of 1d data animated in time.

        Args:
            output_path (str): The place where the plot is saved. If "" or None, the plot is shown in matplotlib.
            dataset_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            axes_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            time_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            dpi (int): The resolution of the frames in dots per inch.
            fps (int): The frames per seconds.
            scale_mode (str): How the scale is changed thorough time. Available methods are:

                * "expand": The y limits increase when needed, but they don't decrease.
                * "adjust_always": Always change the y limits to those of the data.

            latex_label (bool): Whether for use LaTeX code for the plot.

        """
        if output_path:
            ensure_dir_exists(os.path.dirname(output_path))
        axes = self.get_axes(dataset_selector=dataset_selector, axes_selector=axes_selector)
        if len(axes) != 1:
            raise ValueError("Expected 1 axis plot, but %d were provided" % len(axes))
        axis = axes[0]

        gen = self.get_generator(dataset_selector=dataset_selector, axes_selector=axes_selector,
                                 time_selector=time_selector)

        # Set plot labels
        fig, ax = plt.subplots()
        fig.set_tight_layout(True)

        x_name = axis["LONG_NAME"]
        x_units = axis["UNITS"]
        with h5py.File(self.file_list[0], "r") as f:
            y_name = f[self.keys[0]].attrs["LONG_NAME"][0].decode('UTF-8')
            y_units = f[self.keys[0]].attrs["UNITS"][0].decode('UTF-8')
        if latex_label:
            if x_units:
                x_units = "$" + x_units + "$"
            if y_units:
                y_units = "$" + y_units + "$"
            # Names might be text or LaTeX. Try to guess
            if _is_latex(x_name):
                x_name = "$" + x_name + "$"
            if _is_latex(y_name):
                y_name = "$" + y_name + "$"

        if x_units:
            ax.set_xlabel("%s (%s)" % (x_name, x_units))
        else:
            ax.set_xlabel("%s" % (x_name,))
        if y_units:
            ax.set_ylabel("%s (%s)" % (y_name, y_units))
        else:
            ax.set_ylabel("%s" % (y_name,))

        # Plot the points
        x_min, x_max = axis["MIN"], axis["MAX"]
        plot_data, = ax.plot(axis["LIST"], next(gen))
        ax.set_xlim(x_min, x_max)

        time_list = self.get_time_list(time_selector)

        # Prepare a function for the updates
        def update(i):
            """Update the plot, returning the artists which must be redrawn."""
            try:
                new_dataset = next(gen)
            except StopIteration:
                logger.warning("Warning: Tried to add a frame but all data was used")
                return
            label = 't = {0}'.format(time_list[i])
            plot_data.set_ydata(new_dataset[:])
            ax.set_title(label)
            if not scale_mode:
                pass
            elif scale_mode == "expand":
                prev = ax.get_ylim()
                data_limit = [min(new_dataset), max(new_dataset)]
                ax.set_ylim(min(prev[0], data_limit[0]), max(prev[1], data_limit[1]))
            elif scale_mode == "adjust_always":
                ax.set_ylim(min(new_dataset), max(new_dataset))
            return plot_data, ax

        if not output_path:  # "" or None
            # FIXME: Plot in matplotlib window seems not to be working now.
            FuncAnimation(fig, update, frames=np.arange(1, len(time_list) - 1), interval=1)
            plt.show()
        elif output_path.split(".")[-1].lower() == "gif":
            anim = FuncAnimation(fig, update, frames=np.arange(1, len(time_list) - 1), interval=200)
            anim.save(output_path, dpi=dpi, writer='imagemagick')
        else:
            metadata = dict(title=os.path.split(self.data_path)[-1], artist='duat', comment=self.data_path)
            writer = FFMpegWriter(fps=fps, metadata=metadata)
            with writer.saving(fig, output_path, dpi):
                # Iterate over frames
                for i in np.arange(1, len(time_list)):
                    update(i)
                    writer.grab_frame()
                # Keep showing the last frame for the fixed time
                writer.grab_frame()
        plt.close()

    def time_1d_colormap(self, output_path=None, dataset_selector=None, axes_selector=None, time_selector=None,
                         dpi=200, latex_label=True, cmap=None, log_map=False):
        """
        Generate a colormap in an axis and the time.
    
        This function plots a magnitude depending on ONE spatial coordinate (hence the name) and on time as a colormap in
        the cartesian product of such a magnitude and the time.
    
        Args:
            output_path (str): The place where the plot is saved. If "" or None, the plot is shown in matplotlib.
            dataset_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            axes_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            time_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            dpi (int): The resolution of the file in dots per inch.
            latex_label (bool): Whether for use LaTeX code for the plot.
            cmap (str or `matplotlib.colors.Colormap`): The Colormap to use in the plot.
            log_map (bool): Whether the map is plot in log scale.
    
        """
        if output_path:
            ensure_dir_exists(os.path.dirname(output_path))
        axes = self.get_axes(dataset_selector=dataset_selector, axes_selector=axes_selector)
        if len(axes) != 1:
            raise ValueError("Expected 1 axis plot, but %d were provided" % len(axes))
        if len(self.file_list) < 2:
            raise ValueError("Unable to plot a colormap with only one time snapshot")
        axis = axes[0]

        gen = self.get_generator(dataset_selector=dataset_selector, axes_selector=axes_selector,
                                 time_selector=time_selector)

        # Set plot labels
        fig, ax = plt.subplots()

        x_name = axis["LONG_NAME"]
        x_units = axis["UNITS"]
        y_name = "t"
        y_units = self.time_units
        title_name = self.data_name
        title_units = self.units

        if latex_label:
            if x_units:
                x_units = "$" + x_units + "$"
            if y_units:
                y_units = "$" + y_units + "$"
            if title_units:
                title_units = "$" + title_units + "$"
            # Names might be text or LaTeX. Try to guess
            if _is_latex(x_name):
                x_name = "$" + x_name + "$"
            if _is_latex(y_name):
                y_name = "$" + y_name + "$"
            if _is_latex(title_name):
                title_name = "$" + title_name + "$"

        if x_units:
            ax.set_xlabel("%s (%s)" % (x_name, x_units))
        else:
            ax.set_xlabel("%s" % (x_name,))
        if y_units:
            ax.set_ylabel("%s (%s)" % (y_name, y_units))
        else:
            ax.set_ylabel("%s" % (y_name,))

        time_list = self.get_time_list(time_selector)

        # Gather the points
        x_min, x_max = axis["MIN"], axis["MAX"]
        z = np.asarray(list(gen))
        if log_map:
            # Mask manually to prevent a UserWarning
            contour_plot = ax.contourf(axis["LIST"], time_list, np.ma.masked_where(z <= 0, z),
                                       locator=ticker.LogLocator(), cmap=cmap)
        else:
            contour_plot = ax.contourf(axis["LIST"], time_list, z, cmap=cmap)

        ax.set_xlim(x_min, x_max)
        ax.set_ylim(time_list[0], time_list[-1])

        ax.set_title("%s (%s)" % (title_name, title_units))

        fig.colorbar(contour_plot)

        if not output_path:  # "" or None
            plt.show()
        else:
            plt.savefig(output_path, dpi=dpi)

        plt.close()


def get_diagnostic_list(run_dir="."):
    """
    Create a list with the diagnostic found in the given run directory.

    Args:
        run_dir (str): The run directory.

    Returns:
        (list of Diagnostic) List of the diagnostic found.

    """
    diagnostic_list = []
    # This might be not compatible with Windows due to the meaning of \ in patterns, but who cares...
    ms_folder = os.path.join(run_dir, "MS", "")  # Notice the empty string forces the route ending in the separator...
    for root, dirs, files in os.walk(ms_folder):
        if not dirs and files:  # Terminal directory with files in it
            diagnostic_list.append(Diagnostic(root))
    return diagnostic_list


def auto_process(run_dir=".", file_format="mp4", output_dir=None, verbose=None, kwargs_1d=None, num_threads=None):
    """
    Automatically process the files generated by a Osiris run.

    Args:
        run_dir (str): The run directory.
        file_format (str): The preferred format for the generated files (mp4, gif...).
        output_dir (str): Where to output the plots. Default: run_dir/plot.
        verbose (bool): Whether to print to stdout a message when a file is generated.
        kwargs_1d (dict): Additional keywords arguments passed to `time_1d_animation` and `time_dataset_animation`.
        num_threads (int): Number of threads to use to process the files. If None or less than 2, no multiprocessing is
                           used.

    Returns:
        (int) Number of files generated.

    """
    # TODO: Pass kwargs
    # TODO: Choose if prefer video or colormap
    if not output_dir:
        output_dir = os.path.join(run_dir, "plot")
    ensure_dir_exists(output_dir)

    if not kwargs_1d:
        kwargs_1d = {}

    generated = 0

    v_print = print if verbose else lambda *args, **kwargs: None

    if isinstance(num_threads, int) and num_threads < 2:
        num_threads = 0
    if num_threads:
        t = MPCaller(num_threads)

    diagnostics = get_diagnostic_list(run_dir)

    for d in diagnostics:
        filename_base = os.path.join(output_dir, d.data_name)
        d_spatial, d_datasets, d_time = d.shape
        v_print("Generating file(s) for " + d.data_name + "\n- Dimensions: " + str(d.shape))
        if len(d_spatial) == 1 and d_datasets == 1:
            v_print("- Generating: " + filename_base + "." + file_format)
            if num_threads:
                t.add_call(Call(d.time_1d_animation, filename_base + "." + file_format, **kwargs_1d))
            else:
                d.time_1d_animation(filename_base + "." + file_format, **kwargs_1d)
            generated += 1
        elif len(d_spatial) == 1:
            chosen_datasets = [0, 1] if d_datasets == 2 else [0, d_datasets // 2, d_datasets - 1]
            for c in chosen_datasets:
                v_print("- Generating: " + filename_base + "_" + str(c) + "." + file_format)
                if num_threads:
                    t.add_call(
                        Call(d.time_1d_animation, filename_base + "_" + str(c) + "." + file_format, dataset_selector=c,
                             **kwargs_1d))
                else:
                    d.time_1d_animation(filename_base + "_" + str(c) + "." + file_format, dataset_selector=c,
                                        **kwargs_1d)
                generated += 1

    if num_threads:
        t.wait_calls()
    return generated



# -*- coding: UTF-8 -*-
"""Common tools."""

import os
import re
from multiprocessing import Process, Queue
import logging


def ifd(d, v1, v2, v3):
    """
    Conditionally select a value depending on the given dimension number.

    Args:
        d (int): Dimension number (1, 2, or 3).
        v1: The value if d = 1.
        v2: The value if d = 2.
        v3: The value if d = 3.

    Returns:
        v1, v2 or v3.

    """
    if d == 1:
        return v1
    elif d == 2:
        return v2
    elif d == 3:
        return v3
    else:
        raise ValueError("Invalid dimension: %s." % d)


def human_order_key(text):
    """
    Key function to sort in human order.

    """
    # This is based in http://nedbatchelder.com/blog/200712/human_sorting.html
    return [int(c) if c.isdigit() else c for c in re.split('(\d+)', text)]


def ensure_dir_exists(path):
    """
    Ensure a directory exists, creating it if needed.

    Args:
        path: The path to the directory.

    Raises:
        OSError: An error occurred when creating the directory.

    """
    try:
        # Will fail either if exists or unable to create it
        os.makedirs(path)
    except OSError:
        if os.path.exists(path):
            # Directory did [probably] exist
            pass
        else:
            # There was an error on creation, so make sure we know about it
            raise OSError("Unable to create directory " + path)


def get_dir_size(dir_path):
    """Get the size of a directory in bytes."""
    total_size = 0
    for root, dirs, files in os.walk(dir_path):
        for f in files:
            total_size += os.path.getsize(os.path.join(root, f))
    return total_size


def ensure_executable(path, all_users=None):
    """
    Ensure a file is executable.

    Args:
        path (str): the path to the file.
        all_users (bool): whether it should be make executable for the user or for all users.

    """
    st = os.stat(path)
    os.chmod(path, st.st_mode | (0o111 if all_users else 0o100))


def daemonize(func):
    """
    Run a function as a daemon.

    Args:
        func (`Callable`): The function to run as a daemon.

    Raises:
        OSError: An error occurred performing a fork.

    """
    # Create a son
    pid = os.fork()
    if pid > 0:  # Parent
        return
    # We are the son
    os.setsid()

    # Create a grandson

    pid = os.fork()
    if pid > 0:  # Son (not grandson)
        os._exit(os.EX_OK)  # So no cleaning is performed

    # We are the grandson
    func()
    os._exit(os.EX_OK)


class Call:
    """Objectization of a call to be made"""

    def __init__(self, fn, *args, **kwargs):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.fn(*self.args, **self.kwargs)


def _caller(q):
    """Execute calls from a Queue until its value is 'END'."""
    while True:
        data = q.get()
        if data == "END":
            break
        else:
            data()  # Call the supplied argument (lambda o Call instance)


class MPCaller:
    """MultiProcessing Caller. Makes calls using multiple subprocesses."""

    def __init__(self, num_threads=2):
        self.q = Queue()
        self.processes = []
        self.spawn_threads(num_threads)

    def spawn_threads(self, num_threads):
        """Create the required number of threads"""
        for _ in range(num_threads):
            t = Process(target=_caller, args=(self.q,))
            t.daemon = True
            t.start()
            self.processes.append(t)

    def add_call(self, call):
        """Add a call to its stack"""
        self.q.put(call)

    def wait_calls(self):
        """Ask all processes to consume the queue and end.

        After this method is called no threads will remain. Create another instance or call spawn_threads if needed.
        """
        for _ in range(len(self.processes)):
            self.q.put("END")
        for t in self.processes:
            t.join()
        self.processes = []


def tail(path, lines=1, _step=4098):
    """
    Get the last lines of a file.
    
    Args:
        path: Path to the file to read.
        lines: Number of lines to read.
        _step: Size of the step used in the search.

    Returns: (list of str): The lines found.

    """
    # Adapted from glenbot's answer to:
    # http://stackoverflow.com/questions/136168/get-last-n-lines-of-a-file-with-python-similar-to-tail
    f = open(path, "r")
    lines_found = []
    block_counter = -1
    while len(lines_found) < lines:
        try:
            f.seek(block_counter * _step, os.SEEK_END)
        except IOError:  # either file is too small, or too many lines requested
            # read all and exit loop
            f.seek(0)
            lines_found = f.readlines()
            break

        lines_found = f.readlines()

        if len(lines_found) > lines:
            break
        block_counter -= 1

    return lines_found[-lines:]

logging.basicConfig(level=logging.INFO)
logger = logging.Logger("duat")




# -*- coding: UTF-8 -*-
"""Model OSIRIS configuration files."""

import numpy as np

from ..common import ifd, logger


def val_to_fortran(val):
    """
    Transform a value to fortran code.

    Args:
        val: The value to translate.

    Returns: (str) The fortran code.

    """
    # Convert numpy types
    if isinstance(val, np.generic):
        val = val.item()

    t = type(val)
    if t is bool:
        return ".true." if val else ".false."
    elif t is int:
        return str(val)
    elif t is float:
        return ("%.6g" % val).replace("e", "d")
    elif t is str:
        return '"' + val + '"'
    raise TypeError("Unknown type: " + t)


def par_to_fortran(name, val):
    """
    Transform a parameter to fortran code.

    Args:
        name (str): The name of the parameter.
        val: The value of the parameter.

    Returns: A string which assigns the value to the parameter.

    """
    # Convert numpy types
    if isinstance(val, np.generic):
        val = val.item()
    if isinstance(val, np.ndarray):
        val = val.tolist()

    t = type(val)
    if t in [bool, int, float, str]:
        return name + " = " + val_to_fortran(val) + ","
    elif t in [list, tuple]:  # 1D or 2D or 3D

        if type(val[0]) in [list, tuple]:  # 2D or 3D
            if type(val[0][0]) in [list, tuple]:  # 3D
                s = ""
                for i, row in enumerate(val):
                    for j, row_row in enumerate(row):
                        s += name + "(1:" + str(len(row_row)) + ", " + str(i + 1) + ", " + str(
                            j + 1) + ") = " + ", ".join(
                            list(map(val_to_fortran, row_row))) + ",\n  "
                return s
            else:  # 2D
                s = ""
                for i, row in enumerate(val):
                    s += name + "(1:" + str(len(row)) + ", " + str(i + 1) + ") = " + ", ".join(
                        list(map(val_to_fortran, row))) + ",\n  "
                return s
        else:  # 1D
            l = len(val)
            return name + "(1:" + str(l) + ") = " + ", ".join(list(map(val_to_fortran, val))) + ","
    raise TypeError


class MetaSection:
    """Something behaving like a set of sections."""

    def __str__(self):
        return self.to_fortran()

    def to_fortran(self):
        raise NotImplementedError()


class Section(MetaSection):
    """The class defining a configuration block."""

    def __init__(self, name, param=None):
        self.name = name
        if param:
            self.pars = param.copy()
        else:
            self.pars = {}

    def __getitem__(self, ind):
        return self.pars[ind]

    def __setitem__(self, key, value):
        self.pars[key] = value

    def set_par(self, name, val):
        """Add or update the value of a parameter"""
        self.pars[name] = val

    def set_pars(self, **kwargs):
        """Add or update some parameters using keyword arguments"""
        for key in kwargs:
            self.pars[key] = kwargs[key]

    def to_fortran(self):
        s = self.name + "\n{\n"
        for p in self.pars:
            s += "  " + par_to_fortran(p, self.pars[p]) + "\n"
        s += "}\n"
        return s


# TODO: Add a class derived from ConfigSection with prefixed parameters.


class SectionList(MetaSection):
    """
    Class defining a list of sections in a numerical order.

    Here 'section' refers to any subclass of `MetaSection`.
    
    """

    def __init__(self, label=None, lst=None):
        self.label = label if label else ""
        self.lst = lst if lst else []

    def __getitem__(self, ind):
        return self.lst[ind]

    def append_section(self, section):
        self.lst.append(section)

    def to_fortran(self):
        s = ("!---" + self.label + "\n") if self.label else ""
        for sec in self.lst:
            s += sec.to_fortran() + "\n"
        return s


class SectionOrdered(MetaSection):
    """
    Class defining a set of sections that must be outputted in a particular order given by a keyword related to them.

    Here 'section' refers to any subclass of `MetaSection`.
    
    """

    def __init__(self, label=None, order=None, fixed=True, types=None):
        self.label = label if label else ""
        if not order:
            if fixed:
                logger.warning("A ConfigSectionOrdered instance with no order defined cannot be fixed.")
            self.fixed = False
            self.order = []
        else:
            self.fixed = fixed
            self.order = order

        self.types = types if types else {}
        self.subsections = {}

    def __getitem__(self, name):
        if isinstance(name, int):
            name = self.order[name]

        if name in self.subsections:
            return self.subsections[name]
        elif name in self.types:
            self.subsections[name] = self.types[name](name)
            return self.subsections[name]
        else:
            raise ValueError("Subsection %s cannot be implicitly created due to unknown type." % name)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            key = self.order[key]
        self.set_section(key, value)

    def set_section(self, name, section=None):
        """ Add or replace a section."""
        if not section:  # No section was given
            section = Section(name) if "_list" not in name else SectionList(label=name)
        if isinstance(section, str):  # A string was given (abuse of notation)
            section = Section(section) if "_list" not in name else SectionList(label=section)
        if (name in self.subsections) or (name in self.order):
            self.subsections[name] = section
        else:
            if self.fixed:
                raise ValueError("Section %s not expected here. Valid names are %s" % (name, self.order))
            else:
                self.subsections[name] = section
                self.order.append(name)

    def to_fortran(self):
        s = ("!---" + self.label + "\n") if self.label else ""
        for name in self.order:
            if name in self.subsections:
                s += self.subsections[name].to_fortran() + "\n"
        return s


# *****************************************
# Classes describing the semantic structure
# *****************************************


class Species(SectionOrdered):
    """Set of sections defining a species."""

    # Keep the default order as a class variable, but copy to the instance to allow modification
    order = ["species", "profile", "spe_bound", "diag_species"]
    types = {"species": Section, "profile": Section, "spe_bound": Section, "diag_species": Section}

    def __init__(self, d, num=None):
        label = "Configuration for species " + str(num) if num else ""
        SectionOrdered.__init__(self, label=label, order=Species.order, fixed=True, types=Species.types)

        self.set_section("species",
                         Section("species",
                                 {"num_par_max": 2048, "rqm": -1.0, "num_par_x": [2] * d, "vth": [0.1] * 3,
                                        "vfl": [0.0, 0.0, 0.6], "den_min": 1.0e-5, "num_dgam": 0, "dgam": 0}))

        default_profile = ifd(d, {"fx": [[1., 1., 1., 1., 1., 1.]],
                                  "x": [[0., 0.9999, 1.000, 2.000, 2.001, 10000.]]},
                              # 2d
                              {"fx": [[1., 1., 1., 1., 1., 1.], [1., 1., 1., 1., 1., 1.]],
                               "x": [[0., 0.9999, 1.000, 2.000, 2.001, 10000.],
                                     [0., 0.9999, 1.000, 2.000, 2.001, 10000.]]},
                              # 3d
                              {"fx": [[1., 1., 1., 1., 1., 1.], [1., 1., 1., 1., 1., 1.],
                                      [1., 1., 1., 1., 1., 1.]],
                               "x": [[0., 0.9999, 1.000, 2.000, 2.001, 10000.],
                                     [0., 0.9999, 1.000, 2.000, 2.001, 10000.],
                                     [0., 0.9999, 1.000, 2.000, 2.001, 10000.]]}
                              )
        self.set_section("profile", Section("profile", default_profile))

        self.set_section("spe_bound",
                         Section("spe_bound",
                                 {"type": ifd(d, [[0, 0]], [[0, 0], [0, 0]], [[0, 0], [0, 0], [0, 0]])}))

        self.set_section("diag_species", Section("diag_species"))


class ConfigFile(SectionOrdered):
    """
    Set of Sections defining an input file.
    """
    order = ["simulation", "node_conf", "grid", "time_step", "restart", "space", "time", "el_mag_fld", "emf_bound",
             "smooth", "diag_emf", "particles", "species_list", "cathode_list", "neutral_list", "neutral_mov_ions_list",
             "zpulse_list", "current", "smooth_current"]

    # FIXME: There are two "smooths" in the config file definition. I have renamed the second to smooth_current, but the output form must change

    types = {"simulation": Section, "node_conf": Section, "grid": Section, "time_step": Section, "restart": Section,
             "space": Section, "time": Section, "el_mag_fld": Section, "emf_bound": Section, "smooth": Section,
             "diag_emf": Section, "particles": Section, "species_list": SectionList, "cathode_list": SectionList,
             "neutral_list": SectionList, "neutral_mov_ions_list": SectionList, "zpulse_list": SectionList,
             "current": Section, "smooth_current": Section}

    def __init__(self, d):
        """
        Create a default d-dimensional config file.

        Args:
            d(int): The number of dimensions (1, 2 or 3).
            
        """
        SectionOrdered.__init__(self, order=ConfigFile.order, fixed=True, types=ConfigFile.types)

        if d not in [1, 2, 3]:
            raise TypeError("Invalid dimension")

        self["node_conf"] = Section("node_conf", {"node_number": [1] * d, "if_periodic": [True] * d})

        self["grid"] = Section("grid",
                               {"coordinates": "cartesian", "nx_p": ifd(d, [1024], [32, 32], [10, 10, 10])})

        self["time_step"] = Section("time_step", {"dt": ifd(d, 0.07, 0.07, 0.05), "ndump": 10})

        self["space"] = Section("space", {"xmin": [0.0] * d, "xmax": ifd(d, [102.4], [3.2, 3.2], [1.0, 1.0, 1.0]),
                                                "if_move": [False] * d})

        self["time"] = Section("time", {"tmin": 0.0, "tmax": 7.0})

        self["emf_bound"] = Section("emf_bound",
                                    {"type": ifd(d, [[0, 0]], [[0, 0], [0, 0]], [[0, 0], [0, 0], [0, 0]])})

        self["particles"] = Section("particles", {"num_species": 2})

        self["species_list"] = SectionList(label="Species configuration")
        for i in [1, 2]:
            self["species_list"].append_section(Species(d, i))

        # self["pulse_sequence"] = ConfigSection("pulse_sequence", {"num_pulses": 0})

    def write(self, path):
        """
        Save the config file to the specified path.

        Args:
            path: the path of the output file.

        Raises:

        """
        with open(path, "w") as f:
            f.write(self.to_fortran())

    def get_d(self):
        """
        Get the dimension of the configuration filed.

        Returns:
            (int): The dimension according to the mandatory xmax parameter in the space section.
            
        """
        x_max = self["space"]["xmax"]
        return len(x_max) if isinstance(x_max, list) else 1



# -*- coding: UTF-8 -*-
"""Run configuration files with OSIRIS."""


from os import path, remove, walk
from shutil import copyfile
import subprocess
from time import sleep, time
import re

from ..common import ensure_dir_exists, ensure_executable, ifd, tail, logger, get_dir_size

import psutil

# Path to osiris executables - guessed later in the code
osiris_1d = ""
"""Path to the osiris-1D.e file"""
osiris_2d = ""
"""Path to the osiris-2D.e file"""
osiris_3d = ""
"""Path to the osiris-3D.e file"""


def set_osiris_path(folder, warn=True):
    global osiris_1d, osiris_2d, osiris_3d
    if not path.isdir(folder):
        if warn:
            logger.warning("%s is not an existing folder." % folder)
        return

    r = path.join(folder, "osiris-1D.e")
    if path.isfile(r):
        osiris_1d = r
    elif warn:
        logger.warning("osiris-1D not found in %s" % folder)
    r = path.join(folder, "osiris-2D.e")
    if path.isfile(r):
        osiris_2d = r
    elif warn:
        logger.warning("osiris-2D not found in %s" % folder)
    r = path.join(folder, "osiris-3D.e")
    if path.isfile(r):
        osiris_3d = r
    elif warn:
        logger.warning("osiris-3D not found in %s" % folder)


def _find_running_exe(exe):
    """Return the list of the pid of the processes of the argument executable (absolute path)"""
    candidates = []
    for proc in psutil.process_iter():
        try:
            pinfo = proc.as_dict(attrs=['pid', 'exe'])
        except psutil.NoSuchProcess:
            pass
        else:
            if pinfo["exe"] and pinfo['exe'] == exe:
                candidates.append(pinfo['pid'])
    return candidates


class Run:
    """
    An osiris run.
    
    Attributes:
        run_dir (str): Directory where the run takes place.
        total_steps (int): Amount of time steps in the simulation.
        process (psutil.Process): Representation of the process running the simulation.
        
    Notes:
        Only single-process runs are supported at the moment.
        
    """

    def __init__(self, run_dir):
        self.run_dir = run_dir
        # TODO: Handle exceptions
        with open(path.join(run_dir, "os-stdin"), "r") as f:
            text = f.read()
        dt = float(re.match(r".*time_step(.*?){(.*?)dt(.*?)=(.*?),(.*?)}", text, re.DOTALL + re.MULTILINE).group(4))
        tmin = float(re.match(r".*time(.*?){(.*?)tmin(.*?)=(.*?),(.*?)}", text, re.DOTALL + re.MULTILINE).group(4))
        tmax = float(re.match(r".*time(.*?){(.*?)tmax(.*?)=(.*?),(.*?)}", text, re.DOTALL + re.MULTILINE).group(4))

        self.total_steps = int((tmax - tmin) // dt) + 1

        candidates = _find_running_exe(path.join(self.run_dir, "osiris"))

        try:
            if not candidates:  # No process running found
                self.process = None
            elif len(candidates) > 1:
                logger.warning("More than one pid was found for the run.\n"
                              "Multiple processes are not really handled by duat yet, do not trust what you see.")
                self.process = psutil.Process(candidates[0])
            else:
                self.process = psutil.Process(candidates[0])
        except psutil.NoSuchProcess:
            # If the process have died before processing was completed.
            self.process = None

    def __repr__(self):
        if self.is_running():  # Process has not finished yet
            return "Run<%s (%s/%d)>" % (self.run_dir, self.current_step(), self.total_steps)
        else:
            # Badly configured runs also return 0, so do not display the useless return code
            if self.has_error():
                return "Run<%s [FAILED]>" % (self.run_dir,)
            else:
                return "Run<%s>" % (self.run_dir,)

    def current_step(self):
        """
        Find the current simulation step.
        
        Returns: (int) The simulation step or -1 if it could not be found.

        """
        last_line = tail(path.join(self.run_dir, "out.txt"), 8)
        if not last_line:  # Empty file
            return -1
        if re.search("now at  t", last_line[-1]):
            return int(re.match(r".* n = *(.*?)$", last_line[-1]).group(1))
        elif " Osiris run completed normally\n" in last_line:
            return self.total_steps
        else:
            return -1

    def is_running(self):
        """Return True if the simulation is known to be running, or False otherwise."""
        if self.process is None:
            return False
        return self.process.is_running()

    def terminate(self):
        """Terminate the OSIRIS process (if running)."""
        if self.process is not None:
            if self.process.is_running():
                try:
                    self.process.terminate()
                except psutil.NoSuchProcess:
                    # The process has just terminated
                    pass
            else:
                logger.warning("The process had already stopped")
        else:
            logger.warning("Asked for termination of a Run with no known process")

    def kill(self):
        """
        Abruptly terminate the OSIRIS process (if running).
        
        The :func:`~duat.osiris.run.Run.terminate` method should be used instead to perform a cleaner exit.
        """
        if self.process is not None:
            if self.process.is_running():
                try:
                    self.process.kill()
                except psutil.NoSuchProcess:
                    # The process has just terminated
                    pass
            else:
                logger.warning("The process had already stopped")
        else:
            logger.warning("Asked for termination of a Run with no known process")

    def estimated_time(self):
        """
        Estimated time to end the simulation in seconds.
        
        The estimation uses a linear model and considers initialization negligible.
        The modification time of the os-stdin file is used in the calculation. If altered, estimation will be meaningless.
        
        Returns: (float) The estimation of the time to end the simulation or NaN if no estimation could be done.

        """
        if not self.is_running():  # Already finished
            return 0
        else:
            current = self.current_step()
            if current <= 0:  # If not started or error
                return float('nan')
            else:
                elapsed = time() - path.getmtime(path.join(self.run_dir, "os-stdin"))
                return elapsed * (self.total_steps / current - 1)

    def get_size(self):
        """Get the size of all run data in bytes."""
        return get_dir_size(self.run_dir)

    def has_error(self):
        """Search for common error messages in the output file."""
        # TODO: Cache result if reached execution with no error
        with open(path.join(self.run_dir, "out.txt"), "r") as f:
            text = f.read()
        # TODO: Optimize this search
        if "(*error*)" in text or re.search("Error reading .* parameters", text) or re.search("MPI_ABORT was invoked",
                                                                                              text):
            return True
        else:
            return False


def run_config(config, run_dir, prefix=None, clean_dir=True, blocking=None, force=None):
    """
    Initiate a OSIRIS run from a config instance.

    Args:
        config (`ConfigFile`): The instance describing the configuration file.
        run_dir (str): Folder where the run is carried.
        prefix (str): A prefix to run the command (e.g., "qsub", ...).
        clean_dir (bool): Whether to remove the files in the directory before execution.
        blocking (bool): Whether to wait for the run to finish.
        force (str): Set what to do if a running executable is found in the directory. Set to "ignore" to launch anyway,
                     possibly resulting in multiple instances running simultaneously; set to "kill" to terminate the
                     existing processes.

    Returns:
        (tuple): A Run instance describing the execution.

    """
    candidates = _find_running_exe(path.join(run_dir, "osiris"))
    if candidates:
        if force == "ignore":
            logger.warning("Ignored %d running exe found in %s" % (len(candidates), run_dir))
        elif force == "kill":
            logger.warning("Killing %d running exe found in %s" % (len(candidates), run_dir))
            for c in candidates:
                try:
                    psutil.Process(c).terminate()
                except psutil.NoSuchProcess:
                    pass  # If just ended
        else:
            logger.warning("Running exe found in %s. Aborting launch." % run_dir)
            return

    if clean_dir:
        for root, dirs, files in walk(run_dir):
            for f in files:
                remove(path.join(root, f))

        for root, dirs, files in walk(run_dir):
            for f in files:
                logger.warning("Could not remove file %s" % f)

    ensure_dir_exists(run_dir)
    config.write(path.join(run_dir, "os-stdin"))
    osiris_path = path.abspath(path.join(run_dir, "osiris"))
    osiris = ifd(config.get_d(), osiris_1d, osiris_2d, osiris_3d)
    copyfile(osiris, osiris_path)
    ensure_executable(osiris_path)

    if not prefix:  # None or ""
        prefix = ""
    elif prefix[-1] != " ":
        prefix += " "

    proc = subprocess.Popen(prefix + osiris_path + " > out.txt 2> err.txt", shell=True, cwd=path.abspath(run_dir))
    if blocking:
        proc.wait()
    else:  # Sleep a little to check for quickly appearing errors and to allow the shell to start osiris
        sleep(0.2)

    # BEWARE: Perhaps under extreme circumstances, OSIRIS might have not started despite sleeping.
    # This could be solved reinstantiating RUN. Consider it a feature instead of a bug :P

    run = Run(run_dir)

    # Try to detect errors checking the output
    if run.has_error():
        logger.warning(
            "Error detected while launching %s.\nCheck out.txt there for more information or re-run in console." % run_dir)
    return run


# Try to guess the OSIRIS location:
for t in [path.join(path.expanduser("~"), "osiris", "bin"),
          path.join("usr", "local", "osiris", "bin")]:
    set_osiris_path(t, warn=False)
    if osiris_1d and osiris_2d and osiris_3d:
        break

if not (osiris_1d and osiris_2d and osiris_3d):
    if not (osiris_1d or osiris_2d or osiris_3d):
        logger.warning("Warning: no OSIRIS executables were found.")
    else:
        if not osiris_1d:
            logger.warning("Warning: osiris-1D.e not found.")
        if not osiris_2d:
            logger.warning("Warning: osiris-2D.e not found.")
        if not osiris_3d:
            logger.warning("Warning: osiris-3D.e not found.")
        logger.warning("Use the function set_osiris_path or set the variables run.osiris_1d and so to allow the run module to work.")
